package net.knarcraft.hungerarena;

import org.bukkit.entity.Player;

/**
 * A container for all properties stored for a player in an arena
 */
public class ArenaPlayer {

    private Player player;
    private Arena arena;
    private boolean ready;
    private boolean quit;
    private boolean dead;
    private boolean isOut;
    private boolean isWatching;
    private boolean needConfirmation;
    private boolean inArena;
    private boolean frozen;

    /**
     * Instantiates a new arena player
     *
     * @param player           <p>The player this arena player represents</p>
     * @param arena            <p>The arena the player is currently in</p>
     * @param ready            <p>Whether the player is currently ready to start the game</p>
     * @param quit             <p>Whether the player has quit the game</p>
     * @param dead             <p>Whether the player has died</p>
     * @param isOut            <p>Whether the player has left the game</p>
     * @param isWatching       <p>Whether the player is watching the arena</p>
     * @param needConfirmation <p>Whether a conformation is necessary for the player to join the game</p>
     * @param inArena          <p>Whether the player is currently in an arena</p>
     * @param frozen           <p>Whether the player is currently frozen, not allowed to move</p>
     */
    public ArenaPlayer(Player player, Arena arena, boolean ready, boolean quit, boolean dead, boolean isOut,
                       boolean isWatching, boolean needConfirmation, boolean inArena, boolean frozen) {
        this.player = player;
        this.arena = arena;
        this.ready = ready;
        this.quit = quit;
        this.dead = dead;
        this.isOut = isOut;
        this.isWatching = isWatching;
        this.needConfirmation = needConfirmation;
        this.inArena = inArena;
        this.frozen = frozen;
    }

    /**
     * Gets the player this arena player corresponds to
     *
     * @return <p>The actual player</p>
     */
    public Player getPlayer() {
        return player;
    }

    public void setPlayer(Player player) {
        this.player = player;
    }

    /**
     * Gets the arena this arena player is currently in
     *
     * @return <p>The arena this player is in, or null if not in an arena</p>
     */
    public Arena getArena() {
        return arena;
    }

    public void setArena(Arena arena) {
        this.arena = arena;
    }

    /**
     * Gets whether this arena player is ready to start the game
     *
     * @return <p>True if this player is ready</p>
     */
    public boolean isReady() {
        return ready;
    }

    public void setReady(boolean ready) {
        this.ready = ready;
    }

    /**
     * Gets whether this arena player has quit
     *
     * <p>If the player has been kicked, has died or has quit the game, this returns true.</p>
     *
     * @return <p>True if this arena player has quit</p>
     */
    public boolean hasQuit() {
        return quit;
    }

    public void setQuit(boolean quit) {
        this.quit = quit;
    }

    /**
     * Gets whether this arena player has died
     *
     * @return <p>True if this arena player has died</p>
     */
    public boolean isDead() {
        return dead;
    }

    public void setDead(boolean dead) {
        this.dead = dead;
    }

    /**
     * Gets whether this arena player has left the game, and is "out"
     *
     * @return <p>True if this arena player is currently out</p>
     */
    public boolean isOut() {
        return isOut;
    }

    public void setOut(boolean out) {
        isOut = out;
    }

    /**
     * Gets whether this arena player is currently watching the arena
     *
     * @return <p>True if currently watching the arena</p>
     */
    public boolean isWatching() {
        return isWatching;
    }

    public void setWatching(boolean watching) {
        isWatching = watching;
    }

    /**
     * Gets whether ths arena player needs to confirm that they want to join the game
     *
     * <p>As some arenas have entry fees, they might require that a player confirms that they really want to join the
     * arena. If a player hasn't confirmed that they want to join, this will be true.</p>
     *
     * @return <p>True if this arena player needs to confirm that they are joining</p>
     */
    public boolean needConfirmation() {
        return needConfirmation;
    }

    public void setNeedConfirmation(boolean needConfirmation) {
        this.needConfirmation = needConfirmation;
    }

    /**
     * Gets whether this arena player is currently in an arena
     *
     * @return <p>True if in an arena</p>
     */
    public boolean isInArena() {
        return inArena;
    }

    /**
     * Sets whether this arena player is currently in an arena
     *
     * @param inArena <p>True if this arena player is currently in an arena</p>
     */
    public void setInArena(boolean inArena) {
        this.inArena = inArena;
    }

    /**
     * Gets whether this arena player is frozen
     *
     * <p>A frozen player is a player that is currently prevented from moving, such as when players are forced to stay
     * on their pedestals until the countdown expires.</p>
     *
     * @return <p>True if this arena player is frozen</p>
     */
    public boolean isFrozen() {
        return frozen;
    }

    public void setFrozen(boolean frozen) {
        this.frozen = frozen;
    }

}
