package net.knarcraft.hungerarena.Listeners;

import net.knarcraft.hungerarena.Arena;
import net.knarcraft.hungerarena.HungerArena;
import org.bukkit.ChatColor;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.bukkit.block.BlockFace;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockBurnEvent;
import org.bukkit.event.block.BlockFadeEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.event.entity.EntityExplodeEvent;
import org.bukkit.event.player.PlayerBucketEmptyEvent;
import org.bukkit.event.player.PlayerBucketFillEvent;

import java.util.List;
import java.util.UUID;

/**
 * A listener for any unwanted placing of or destruction of blocks
 */
public class BlockChangeListener implements Listener {
    public final HungerArena plugin;

    /**
     * Instantiates a new block change listener
     *
     * @param hungerArena <p>A reference to the hungerArena plugin</p>
     */
    public BlockChangeListener(HungerArena hungerArena) {
        this.plugin = hungerArena;
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onBlockBreak(BlockBreakEvent event) {
        //If already cancelled, abort
        if (event.isCancelled()) {
            return;
        }

        Block block = event.getBlock();
        Player player = event.getPlayer();
        UUID playerId = player.getUniqueId();

        //If the player has the HungerArena.arena permission, allow bypassing the always protect protection
        boolean alwaysProtect = false;
        if (!player.hasPermission("HungerArena.arena")) {
            alwaysProtect = alwaysProtect();
        }

        //Find the id of the arena the player is currently in
        int arenaId = getArenaId(player, alwaysProtect);

        //If the player cannot be found in the arena, abort unless always protect is enabled
        if (!plugin.inArena.get(arenaId).contains(playerId) && !alwaysProtect) {
            return;
        }

        if (plugin.config.getBoolean("Protected_Arena")) {
            //Prevent any block breakage
            event.setCancelled(true);
            player.sendMessage(ChatColor.RED + "You can't break blocks while playing!");
        } else if ((joinAble(arenaId) || alwaysProtect) && (!plugin.restrictedWorlds ||
                plugin.worldNames.containsValue(player.getWorld().getName()))) {
            List<String> whitelist = plugin.management.getStringList("blocks.whitelist");
            boolean emptyWhitelist = whitelist.isEmpty();
            boolean useAsBlacklist = plugin.management.getBoolean("blocks.useWhitelistAsBlacklist");
            //If not in whitelist
            if ((emptyWhitelist || !whitelist.contains(block.getType().name())) ^ useAsBlacklist) {
                //If not in whitelist, or in whitelist, but treated as blacklist, deny breaking the block
                event.setCancelled(true);
                player.sendMessage(ChatColor.RED + "That is an illegal block!");
            } else {
                //Store the removed block to restore it later
                addDestroyedBlockToList(block, block.getWorld().getName(), arenaId);
            }
        }
    }

    @EventHandler(priority = EventPriority.LOWEST)
    public void onExplosion(EntityExplodeEvent event) {
        //If already cancelled, abort
        if (event.isCancelled()) {
            return;
        }

        List<Block> destroyedBlocks = event.blockList();
        Entity explosionCause = event.getEntity();
        String explosionWorld = explosionCause.getWorld().getName();
        for (int arenaId : plugin.arenas.keySet()) {
            registerArenaExplosion(arenaId, destroyedBlocks, explosionCause, explosionWorld);
        }
    }

    /**
     * Registers an explosion in the given arena
     *
     * @param arenaId         <p>The arena id to register the explosion for</p>
     * @param destroyedBlocks <p>The blocks destroyed in the explosion</p>
     * @param explosionCause  <p>The cause of the explosion</p>
     * @param explosionWorld  <p>The world in which the explosion happened</p>
     */
    private void registerArenaExplosion(int arenaId, List<Block> destroyedBlocks, Entity explosionCause,
                                        String explosionWorld) {
        //If not active (join-able) and always protect is disabled, do nothing
        if (!joinAble(arenaId) && !alwaysProtect()) {
            return;
        }

        if (isDifferentWorld(arenaId, explosionWorld)) {
            return;
        }

        if (explosionCause.getType() == EntityType.PRIMED_TNT) {
            //Put the destroyed TNT back to where it was, save it to destroyed blocks, and remove it again
            Location explosionLocation = explosionCause.getLocation();
            explosionLocation.getBlock().setType(Material.TNT);
            Block tnt = explosionLocation.getBlock();
            addDestroyedBlockToList(tnt, explosionWorld, arenaId);
            tnt.setType(Material.AIR);
        }
        //Register all destroyed blocks, except air
        for (Block destroyedBlock : destroyedBlocks) {
            if (!destroyedBlock.getType().isAir()) {
                addDestroyedBlockToList(destroyedBlock, explosionWorld, arenaId);
            }
        }

    }

    @EventHandler(priority = EventPriority.LOWEST)
    public void onBlockBurn(BlockBurnEvent event) {
        //If already cancelled, do nothing
        if (event.isCancelled()) {
            return;
        }

        Block burnedBlock = event.getBlock();
        String blockWorldName = burnedBlock.getWorld().getName();
        for (int arenaId : plugin.arenas.keySet()) {
            //If not join-able, and always protect isn't enabled, do nothing
            if (!joinAble(arenaId) && !alwaysProtect()) {
                continue;
            }

            //If restricted to HungerArena worlds, and the world is unset, or not the same, do nothing
            if (isDifferentWorld(arenaId, blockWorldName)) {
                continue;
            }

            //Register the burned block
            addDestroyedBlockToList(burnedBlock, blockWorldName, arenaId);
        }
    }

    /**
     * Gets whether the given arena is in a different world than the given world name
     *
     * @param arenaId   <p>The name of the arena to check</p>
     * @param worldName <p>The name of the world to check</p>
     * @return <p>True if the arena is not in the given world</p>
     */
    private boolean isDifferentWorld(int arenaId, String worldName) {
        String arenaWorldName = plugin.worldNames.get(arenaId);
        //If restricted to HungerArena worlds, and the world is unset, or not the same, do nothing
        return plugin.restrictedWorlds && (arenaWorldName == null ||
                !arenaWorldName.equalsIgnoreCase(worldName));
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void blockPlace(BlockPlaceEvent event) {
        //If the event is already cancelled, do nothing
        if (event.isCancelled()) {
            return;
        }

        Block placedBlock = event.getBlock();
        Player player = event.getPlayer();

        boolean alwaysProtect = false;
        if (!player.hasPermission("HungerArena.arena")) {
            alwaysProtect = alwaysProtect();
        }

        //Find the id of the arena the player is currently in
        int arenaId = getArenaId(player, alwaysProtect);

        //If the player cannot be found in the arena, abort unless always protect is enabled
        if (!plugin.inArena.get(arenaId).contains(player.getUniqueId()) && !alwaysProtect) {
            return;
        }

        //If not join-able, or always protect is disabled, do nothing
        if (!joinAble(arenaId) && !alwaysProtect) {
            return;
        }

        //If restricted to HungerArena worlds, and not a registered world, do nothing
        if (plugin.restrictedWorlds && !plugin.worldNames.containsValue(placedBlock.getWorld().getName())) {
            return;
        }

        Block blockBeneath = placedBlock.getRelative(BlockFace.DOWN);
        World world = placedBlock.getWorld();
        if (placedBlock.getType().hasGravity() && !blockBeneath.getType().isSolid()) {
            //If affected by gravity, and not placed on a solid surface, find the position the block will fall to
            Block finalPosition = findFinalGravityBlockPosition(placedBlock, world, player);
            if (finalPosition != null) {
                addPlacedBlockToList(finalPosition, world.getName(), arenaId);
            }
        } else if (!placedBlock.getType().hasGravity()) {
            addPlacedBlockToList(placedBlock, world.getName(), arenaId);
        }
    }

    /**
     * Finds the final position of a placed block affected by gravity
     *
     * @param placedBlock <p>The placed block</p>
     * @param world       <p>The world the block was placed in</p>
     * @param player      <p>The player that placed the block</p>
     * @return <p>The final position the block will end up in</p>
     */
    private Block findFinalGravityBlockPosition(Block placedBlock, World world, Player player) {
        int x = placedBlock.getX();
        int z = placedBlock.getZ();
        int y = placedBlock.getY() - 1;
        Block blockBeneath = world.getBlockAt(x, y, z);
        //TODO: This doesn't really account for sand placed on torches, where the sand would disappear

        while (!blockBeneath.getType().isSolid()) {
            blockBeneath = world.getBlockAt(x, --y, z);
            player.sendMessage(blockBeneath.getType().toString().toLowerCase());
            if (blockBeneath.getType().isSolid()) {
                return world.getBlockAt(x, y + 1, z);
            }
        }
        return null;
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onBucketEmpty(PlayerBucketEmptyEvent event) {
        //If already cancelled, do nothing
        if (event.isCancelled()) {
            return;
        }

        //If not in an arena, do nothing
        if (plugin.getArena(event.getPlayer()) == null) {
            return;
        }
        //TODO: Shouldn't this also follow protect always? And there is no actual prevention of placement

        int arenaId = plugin.getArena(event.getPlayer()).getArenaId();

        //If not currently join-able, don't do anything
        if (!joinAble(arenaId)) {
            return;
        }

        //If in a protected world, register the placed bucket
        if (!plugin.restrictedWorlds || plugin.worldNames.containsValue(event.getPlayer().getWorld().getName())) {
            Block clickedBlock = event.getBlockClicked().getRelative(event.getBlockFace());
            String world = clickedBlock.getWorld().getName();
            addPlacedBlockToList(clickedBlock, world, arenaId);
        }
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onBucketFill(PlayerBucketFillEvent event) {
        //If already cancelled, do nothing
        if (event.isCancelled()) {
            return;
        }

        //If not in an arena, do nothing
        if (plugin.getArena(event.getPlayer()) == null) {
            return;
        }

        int arena = plugin.getArena(event.getPlayer()).getArenaId();

        boolean isPlaying = plugin.inArena.get(arena).contains(event.getPlayer().getUniqueId());
        boolean enabledInWorld = plugin.worldNames.containsValue(event.getPlayer().getWorld().getName());

        if (joinAble(arena) && (isPlaying && (!plugin.restrictedWorlds || enabledInWorld))) {
            Block block = event.getBlockClicked().getRelative(event.getBlockFace());
            String worldName = block.getWorld().getName();
            addDestroyedBlockToList(block, worldName, arena);
        }

    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void blockMelt(BlockFadeEvent event) {
        if (event.isCancelled()) {
            return;
        }

        for (int arena : plugin.canJoin.keySet()) {
            boolean isInWorld = plugin.worldNames.get(arena).equalsIgnoreCase(event.getBlock().getWorld().getName());
            if (joinAble(arena) || alwaysProtect() &&
                    (!plugin.restrictedWorlds || plugin.worldNames.get(arena) != null && isInWorld)) {
                Block block = event.getBlock();
                String worldName = block.getWorld().getName();
                Material material = block.getType();
                if (material == Material.FIRE || material.isAir()) {
                    continue;
                }
                addDestroyedBlockToList(block, worldName, arena);
            }
        }
    }

    //SubRoutines:
    private boolean alwaysProtect() {
        return plugin.config.getBoolean("Protected_Arena_Always");
    }

    /**
     * Adds the given block to the list of destroyed blocks for the given arena
     *
     * @param block     <p>The block removed</p>
     * @param worldName <p>The world the block was destroyed in</p>
     * @param arena     <p>The arena affected</p>
     */
    private void addDestroyedBlockToList(Block block, String worldName, int arena) {
        int x = block.getX();
        int y = block.getY();
        int z = block.getZ();
        String blockTypeName = block.getType().name();
        String blockDataString;
        String splitter = ";";
        String direction = "";
        block.getState().getBlockData();
        blockDataString = block.getBlockData().getAsString();
        String coordinates = worldName + splitter + x + splitter + y + splitter + z + splitter + blockTypeName +
                splitter + blockDataString + splitter + arena + direction;
        List<String> blocks = plugin.data.getStringList("Blocks_Destroyed");
        if (!plugin.data.getStringList("Blocks_Placed").contains(String.format("%s,%d,%d,%d,%s", worldName, x, y,
                z, arena))) {
            blocks.add(coordinates);
            plugin.data.set("Blocks_Destroyed", blocks);
            plugin.saveData();
        }
    }

    /**
     * Adds the given block to the list of placed blocks for the given arena
     *
     * @param block     <p>The block placed</p>
     * @param worldName <p>The world the block was placed in</p>
     * @param arena     <p>The arena affected</p>
     */
    private void addPlacedBlockToList(Block block, String worldName, int arena) {
        String coordinates = String.format("%s,%d,%d,%d,%d", worldName, block.getX(), block.getY(), block.getZ(), arena);
        List<String> blocks = plugin.data.getStringList("Blocks_Placed");
        if (!blocks.contains(coordinates)) {
            blocks.add(coordinates);
            plugin.data.set("Blocks_Placed", blocks);
            plugin.saveData();
        }
    }

    /**
     * Gets whether the given arena can be joined
     *
     * @param arena <p>The arena to check</p>
     * @return <p>True if the arena can be joined</p>
     */
    private boolean joinAble(int arena) {
        return plugin.arenas.get(arena).isJoinAble();
    }

    /**
     * Gets the id of the given player's current arena
     *
     * @param player        <p>The player to get the id for</p>
     * @param alwaysProtect <p>Whether always protect is enabled</p>
     * @return <p>The id of the arena the player is in</p>
     */
    private int getArenaId(Player player, boolean alwaysProtect) {
        int arenaId = 1;
        Arena pluginArena = plugin.getArena(player);
        if (pluginArena != null) {
            arenaId = pluginArena.getArenaId();
        } else if (alwaysProtect) {
            //If always protect is enabled, choose the first arena in the player's world, as an arena is necessary
            String playerWorld = player.getWorld().getName();
            for (int worldNameIndex : plugin.worldNames.keySet()) {
                String worldName = plugin.worldNames.get(worldNameIndex);
                if (worldName != null && worldName.equals(playerWorld)) {
                    arenaId = worldNameIndex;
                }
            }
        }
        return arenaId;
    }

}